#include "bnetfwd.hh"
#include "logger.hh"
#include <iostream>
#include <algorithm>
using namespace std;

#define ALL_IPS_EXTERNAL

/**************************************
 * bnet_tcp_proxied_connection::tcp_state
 */

bool bnet_tcp_proxied_connection::tcp_state::new_packet(const char *head)
{
	memcpy(&pkt_head, head, BNET_HEAD_LEN);
	if (pkt_head.len < BNET_HEAD_LEN)
		return false;
	packet_bytes_left = pkt_head.len -= BNET_HEAD_LEN;
	pkt_offs = subst_len = 0;
	subst_offs = pkt_head.len + 1;
	subst_data.clear();
	pkt_handler = NULL;
	return true;
}

bool bnet_tcp_proxied_connection::tcp_state::subst(virtual_buffer<char>& data, buffered_io& io)
{
	if (subst_offs > pkt_offs) {
		size_t diff = subst_offs - pkt_offs;
		size_t len = data.count_elements_left();
		io.out_append_and_move(data, diff);
		len -= data.count_elements_left();
		if (DEBUG(DATASUBST))
			logger	<< timestamp << " subst on [" << io.fileno() << "]: passed " << len
				<< '/' << diff << " @ " << pkt_offs << endl;
		pkt_offs += len;
	}

	if (subst_offs == pkt_offs) {
		if (subst_len) {
			size_t len = data.count_elements_left();
			data.advance(subst_len);
			len -= data.count_elements_left();
			if (DEBUG(DATASUBST))
				logger	<< timestamp << " subst on [" << io.fileno() << "]: skipped " << len
					<< '/' << subst_len << " @ " << pkt_offs << endl;
			subst_len -= len;
			subst_offs += len;
			pkt_offs += len;
		}
		if (subst_data.count_elements_left()) {
			size_t len = subst_data.count_elements_left();
			io.out_append_and_move(subst_data, len);
			if (DEBUG(DATASUBST))
				logger	<< timestamp << " subst on [" << io.fileno() << "]: inserted "
					<< (len - subst_data.count_elements_left())
					<< '/' << len << " @ " << pkt_offs << endl;
			if (!subst_data.count_elements_left()) {
				subst_offs = pkt_head.len + 1;
				return true;
			} else {
				logger	<< timestamp << " subst on [" << io.fileno() << "]: **BUG**: "
					<< subst_data.count_elements_left() << " subst data left after append" << endl;
			}
		}
	}

	return false;
}

/**************************************
 * bnet_tcp_proxied_connection
 */

void bnet_tcp_proxied_connection::break_packet(tcp_forward::forwarding_side *_slave)
{
	bool server_side = (_slave == slave[TO_SERVER]);
	tcp_forward::forwarding_side *sibling_slave = slave[!server_side];
	tcp_state& slave_state = (server_side ? server_state : client_state);
	virtual_buffer<char> part;
	size_t blocklen = 0;
	size_t more = 0;

	if (slave_state.packet_bytes_left > 0) {
		part = virtual_buffer<char>(_slave->inbuf, slave_state.packet_bytes_left);
		blocklen = part.count_elements_left();

		if (DEBUG(BNETTCPDATA) && want_packet_dump(slave_state.pkt_head.type))
			logger	<< timestamp << " data fragment [" << _slave->fileno() << "]: " << part;

		if (slave_state.pkt_handler)
			more = (this->*slave_state.pkt_handler)(part);
		else if (slave_state.subst(part, *sibling_slave) && part.count_elements_left())
			slave_state.subst(part, *sibling_slave);

		blocklen -= part.count_elements_left();
		_slave->inbuf.advance(blocklen);
		slave_state.packet_bytes_left -= blocklen;
	} else {
		if (!_slave->in_get_block(part, BNET_HEAD_LEN))
			return;

		bool ok = slave_state.new_packet(part.get_head_ptr());
		struct _bnet_packet_head& bh = slave_state.pkt_head;
		size_t bh_len = bh.len;
		if (ok)
			bh_len += BNET_HEAD_LEN;

		if ((DEBUG(BNETTCPFLOW) || DEBUG(BNETTCPDATA)) && want_packet_dump(bh.type)) {
			logger	<< timestamp << " [" << _slave->fileno() << "]: got packet header: 0x"
				<< hex << setw(4) << setfill('0') << bh.type << dec << " ("
				<< bnet_packet_type_str(bh.type) << "), len " << bh_len << endl;
		}

		if (!ok) {
			logger	<< timestamp << " ignoring forward [" << fileno(0) << ',' << fileno(1)
				<< "]: protocol error: len < " << BNET_HEAD_LEN << endl;
			state = TRANSPARENT;
			return;
		}

		if (server_side)
			more = handle_header_from_server_cb(part);
		else	more = handle_header_from_client_cb(part);
	}

	if (more > slave_state.packet_bytes_left) {
		logger	<< timestamp << " ignoring forward [" << fileno(0) << ',' << fileno(1)
			<< "]: protocol error: short packet; left " << slave_state.packet_bytes_left << '/' << more << endl;
		state = TRANSPARENT;
		return;
	}

	virtual_buffer<char> tmp;
	if (more && _slave->in_get_block(tmp, more)) {
		logger	<< timestamp << " ignoring forward [" << fileno(0) << ',' << fileno(1)
			<< "]: **BUG** in_get_block() unexpectedly returned true" << endl;
		state = TRANSPARENT;
		slave[_slave == slave[0]]->out_append(tmp);
		return;
	}
}

void bnet_tcp_proxied_connection::slave_read_callback(tcp_forward::forwarding_side *_slave)
{
	switch (state) {
	case VALID:
		break_packet(_slave);
		break;

	case TRANSPARENT:
		tcp_forward::slave_read_callback(_slave);
		break;
	
	case INIT:
		check_connection_type(_slave);
		break;
	}
}

void bnet_tcp_proxied_connection::check_connection_type(tcp_forward::forwarding_side *_slave)
{
	int t = (unsigned char)*_slave->inbuf.get_head_ptr();
	if (_slave != slave[TO_CLIENT]) {
		logger	<< timestamp << " ignoring forward [" << fileno(0) << ',' << fileno(1)
			<< "]: server talked first" << endl;
		state = TRANSPARENT;
	} else if (t != 1) {
		logger	<< timestamp << " ignoring forward [" << fileno(0) << ',' << fileno(1)
			<< "]: conntype " << t << " != 1" << endl;
		state = TRANSPARENT;
	} else {
		logger	<< timestamp << " forward [" << fileno(0) << ',' << fileno(1)
			<< "] is valid" << endl;
		state = VALID;
		slave[TO_SERVER]->out_append_and_move(_slave->inbuf, 1);
	}
}

void bnet_tcp_proxied_connection::slave_error_callback(tcp_forward::forwarding_side *_slave, int err)
{
	master->connection_error(this, _slave == slave[TO_SERVER], err);
}

size_t bnet_tcp_proxied_connection::handle_header_from_client_cb(virtual_buffer<char>& data)
{
	if (client_state.pkt_head.type == BNET_PT_HELLO)
		setup_client_hello();
	get_server_io()->out_append_and_move(data, data.count_elements_left());

	return 0;
}

size_t bnet_tcp_proxied_connection::handle_header_from_server_cb(virtual_buffer<char>& data)
{
	switch (server_state.pkt_head.type) {
	case BNET_PT_HELLO:
		server_state.pkt_handler = &bnet_tcp_proxied_connection::process_server_hello;
		break;

	case BNET_PT_GAMEINFO:
		server_state.pkt_handler = &bnet_tcp_proxied_connection::process_server_gameinfo_get_count;
		break;
	}

	get_client_io()->out_append_and_move(data, data.count_elements_left());

	return 0;
}

void bnet_tcp_proxied_connection::setup_client_hello()
{
	socket_address me;
	get_server_io()->get_my_address(me);
	uint32_t myip = htonl(me.get_inet_addr());
	client_state.subst_data.append((char *)&myip, 4);
	client_state.subst_offs = 0x14;
	client_state.subst_len = 4;
}

size_t bnet_tcp_proxied_connection::process_server_hello(virtual_buffer<char>& data)
{
	if (data.count_elements_left() < 12)
		return 12;

	copy(data.get_head_ptr() + 4, data.get_head_ptr() + 12, (char *)&challenge);
	get_client_io()->out_append_and_move(data, 12);
	server_state.pkt_offs += 12;
	server_state.pkt_handler = NULL;
	got_challenge = true;

	if (DEBUG(DATAFORMAT)) {
		logger	<< timestamp << " server challenge:" << hex << setfill('0');
		unsigned char *p = (unsigned char *)&challenge;
		for(int i = 0; i < sizeof(challenge); ++i)
			logger << ' ' << setw(2) << (unsigned)*(p++);
		logger	<< dec << setfill(' ') << endl;
	}

	return 0;
}

size_t bnet_tcp_proxied_connection::process_server_gameinfo_get_count(virtual_buffer<char>& data)
{
	if (data.count_elements_left() < 4)
		return 4;

	uint32_t n = *(uint32_t *)data.get_head_ptr();
	virtual_buffer<char> tmp(data, 4);
	server_state.subst(tmp, *get_client_io());
	data.advance(4);

	logger	<< timestamp << " [" << fileno(1) << "]: " << n << " games" << endl;
/*
	for(gamevector::iterator p = games.begin(); p != games.end(); ++p)
		if (p->peer) {
			p->peer->drop();
			p->peer = NULL;
		}

	games.resize(n);
*/	cur_game = 0;
	gamelist_state = STRUCT;

	if (n)
		server_state.pkt_handler = &bnet_tcp_proxied_connection::process_server_gameinfo;
	return 0;
}

size_t bnet_tcp_proxied_connection::process_server_gameinfo(virtual_buffer<char>& data)
{
	if (gamelist_state == STRUCT) {
		if (data.count_elements_left() < BNET_GAMEINFO_LEN)
			return BNET_GAMEINFO_LEN;

		_bnet_gameinfo_struct game = *(_bnet_gameinfo_struct *)data.get_head_ptr();

		socket_address::ipportpair addr(ntohl(game.ip), ntohs(game.port));
		socket_address::ipportpair mask = master->map_host_for_local(addr);

		game.ip = htonl(mask.ip);
		game.port = htons(mask.port);
		server_state.subst_offs = server_state.pkt_offs;
		server_state.subst_len = BNET_GAMEINFO_LEN;
		server_state.subst_data.append((char *)&game, BNET_GAMEINFO_LEN);

		logger	<< timestamp << ' ' << cur_game << " masquerading " << addr << " as " << mask << endl;

		gamelist_state = GAMENAME;
	} else if (gamelist_state == GAMENAME || gamelist_state == PADSTRZ || gamelist_state == MAPNAME) {
		size_t n = data.count_elements_left();
		char *p = data.get_head_ptr();
		while (n > 0 && *p)
			++p, --n;
		if (!*p) {
			++p, --n;
			switch (gamelist_state) {
				case GAMENAME:	gamelist_state = PADSTRZ; break;
				case PADSTRZ:	gamelist_state = MAPNAME; break;
				case MAPNAME:	gamelist_state = STRUCT; ++cur_game; break;
				default:	abort();
			}
		}

		n = data.count_elements_left() - n;
		server_state.subst_offs = server_state.pkt_offs + n;

		if (DEBUG(DATAFORMAT))
			logger	<< timestamp << " string[" << n << ']' << endl;
	} else
		abort();

	server_state.subst(data, *get_client_io());
	return 0;
}

/**************************************
 * bnet_tcp_forward_listener
 */

tcp_forward *bnet_tcp_forward_listener::create_forward()
{
	return new bnet_tcp_proxied_connection(master);
}

void bnet_tcp_forward_listener::connected(tcp_forward *fwd)
{
	master->new_client(static_cast<bnet_tcp_proxied_connection *>(fwd));
}

void bnet_tcp_forward_listener::connect_died(int ofd, int cfd, int err)
{
	logger << timestamp << " connect [" << ofd << ',' << cfd << "] died: " << errno_str(err) << endl;
}

bool bnet_tcp_forward_listener::validate_connection(int ofd, int cfd)
{
	socket_address from;
	io_socket(ofd, false).get_peer_address(from);
	return master->validate_connection(from);
}

/**************************************
 * bnet_udp_peer
 */

bnet_udp_peer::bnet_udp_peer(bnet_udp_forwarder *mx, const socket_address& peer, bnet_udp_forwarder *send_via)
	: udp_mx_peer(mx, peer), via(send_via)
{
}

void bnet_udp_peer::mangle_player_packet(struct _bnet_udp_packet__player *data)
{
	if (data->id != 0)
		return;
	if (data->len <= sizeof(*data) - 4)
		return;
	if (data->w1[2] != 0x600)
		return;
	if (data->addr_type != 2)
		return;
	if (data->pad[0] != 0)
		return;
	if (data->pad[1] != 0)
		return;

	socket_address::ipportpair ipp(ntohl(data->ip), ntohs(data->port));
	bnet_udp_forwarder *mx = static_cast<bnet_udp_forwarder *>(this->mx);
	ipp = mx->master->map_host_for_addr(ipp, mx->get_forward_target());
	data->ip = htonl(ipp.ip);
	data->port = htons(ipp.port);
}

void bnet_udp_peer::pkt_read()
{
	bnet_udp_forwarder *mx = static_cast<bnet_udp_forwarder *>(this->mx);

	if (DEBUG(BNETUDPFLOW))
		logger	<< timestamp << " got packet [" << mx->fileno() << "] from " << mx->packet_from
			<< " forwarding [" << via->fileno() << "] as " << via->get_my_address()
			<< " to " << mx->get_forward_target() << endl;
	if (DEBUG(BNETUDPDATA))
		logger	<< timestamp << " packet [" << mx->fileno() << "]: " << mx->packet;
#ifndef ALL_IPS_EXTERNAL
	struct _bnet_udp_packet__player *data = (struct _bnet_udp_packet__player *)mx->packet.get_head_ptr();
	if (mx->packet.count_elements_left() > sizeof(*data))
		mangle_player_packet(data);
#endif
	via->out_add_to(mx->packet, mx->get_forward_target());
}

/**************************************
 * bnet_udp_forwarder
 */

udp_mx_peer *bnet_udp_forwarder::create_peer()
{
	return (master->*create_func)(this, packet_from);
}

bool bnet_udp_forwarder::validate_new_peer()
{
	if (accept_all_peers)
		return true;

	return master->validate_peer(packet_from);
}

bnet_udp_forwarder::~bnet_udp_forwarder()
{
	socket_address::ipportpair target(forward_target);

	while (!peers.empty()) {
		bnet_udp_peer *peer = static_cast<bnet_udp_peer *>(peers.begin()->second);

		bnet_udp_peer *target_peer = static_cast<bnet_udp_peer *>(peer->via->find_peer(target));
		if (target_peer) {
			if (target_peer->via != this)
				logger	<< timestamp << "target_peer->via != me **BUG** while destroying UDP "
					<< target << " mapping" << endl;

			target_peer->drop();
		} else {
			logger	<< timestamp << "target_peer not found **BUG** while destroying UDP "
				<< target << " mapping" << endl;
		}
		peer->drop();
	}
}

/**************************************
 * ip_semaphores
 */

void ip_semaphores::put(uint32_t ip)
{
	ip_set_type::iterator p = set.find(ip);

	if (p != set.end())
		++p->second;
	else
		set.insert(ip_set_type::value_type(ip, 1));
}

bool ip_semaphores::get(uint32_t ip)
{
	ip_set_type::iterator p = set.find(ip);

	if (p == set.end())
		return false;

	if (!--p->second)
		set.erase(p);
	
	return true;
}

/**************************************
 * bnet_forwarder
 */

bnet_forwarder::bnet_forwarder(const socket_address& _listen, const socket_address& _server)
	: tcp_listener(this, _server), udp_listener(this, _server, &bnet_forwarder::new_udp_client, false), server(_server)
{
	external_udp = new udp_mx_map;
	external_udp->create_func = &bnet_forwarder::new_udp_gamer;
	address_detector(PF_INET).execute(_server, external_udp->bind_address);
	external_udp->bind_address.set_inet_port(0);
	external_udp->mark_str = "external IF";

#ifndef ALL_IPS_EXTERNAL
	internal_udp->create_func = &bnet_forwarder::new_udp_client;
	internal_udp->bind_address = _listen;
	internal_udp->bind_address.set_inet_port(0);
	internal_udp->mark_str = "internal IF";
#else
	internal_udp = external_udp;
#endif

	logger	<< timestamp << " internal binding: " << internal_udp->bind_address
		<< "; external binding: " << external_udp->bind_address << endl;
	logger	<< timestamp << " proxying " << _listen << " to " << _server << endl;
	tcp_listener.bind(_listen);
	tcp_listener.listen();
	udp_listener.bind(_listen);
	udp_listener.start();
	logger	<< timestamp << " listeners: TCP [" << tcp_listener.fileno()
		<< "], UDP [" << udp_listener.fileno() << ']' << endl;
}

bnet_forwarder::~bnet_forwarder()
{
#ifndef ALL_IPS_EXTERNAL
	delete internal_udp;
#endif
	delete external_udp;
}

void bnet_forwarder::new_client(bnet_tcp_proxied_connection *client)
{
	socket_address from;
	io_socket(client->fileno(0), false).get_peer_address(from);
	clients.insert(client);
	allowed_udp_ips.put(from.get_inet_addr());
}

bool bnet_forwarder::validate_connection(const socket_address& from)
{
	return true;
}

void bnet_forwarder::connection_error(bnet_tcp_proxied_connection *client, bool server_side, int err)
{
	logger	<< timestamp << " forward [" << client->fileno(0) << ',' << client->fileno(1)
		<< "] died on " << (server_side ? "server" : "client") << ": " << errno_str(err) << endl;

	clients.erase(client);
	delete client;
	throw err;
}

bool bnet_forwarder::validate_peer(const socket_address& from)
{
	return allowed_udp_ips.get(from.get_inet_addr());
}

bnet_udp_forwarder *bnet_forwarder::create_udp_masking_port(udp_mx_map& udp_map, const socket_address& addr_masked)
{
	/* znalezienie obrazu hosta (portu na "przeciwległym" adresie) */
	udp_mx_map::iterator p = udp_map.find(addr_masked);
	if (p == udp_map.end()) {
		/* utworzenie obrazu hosta */
		bnet_udp_forwarder *maskport = new bnet_udp_forwarder(this, addr_masked, udp_map.create_func, true);
		/* powiązanie host -> obraz */
		p = udp_map.insert(udp_mx_map::value_type(addr_masked, maskport)).first;
		/* podpięcie do adresu */
		maskport->bind(udp_map.bind_address);
		/* włączenie przekazywania pakietów */
		maskport->start();
	}

	if (DEBUG(UDPMASKING))
		logger	<< timestamp << " UDP " << addr_masked << " is masked as ["
			<< p->second->fileno() << "] " << p->second->get_my_address()
			<< " (" << udp_map.mark_str << ')' << endl;

	return p->second;
}

udp_mx_peer *bnet_forwarder::new_udp_peer(udp_mx_map& udp_map, bnet_udp_forwarder *mx, const socket_address& from)
{
	/* znalezienie obrazu klienta na zewnętrznym adresie */
	bnet_udp_forwarder *maskport = create_udp_masking_port(udp_map, from);

	if (DEBUG(UDPMASKING))
		logger	<< timestamp << " new UDP forward: " << from << " (["
			<< maskport->fileno() << "] " << maskport->get_my_address()
			<< ") <-> ([" << mx->fileno() << "] " << mx->get_my_address() << ") "
			<< mx->get_forward_target() << endl;

	/* dodanie przekierowania serwer -> klient */
	maskport->add_peer(new bnet_udp_peer(maskport, mx->get_forward_target(), mx));
	/* zwrócenie przekierowania klient -> serwer */
	return new bnet_udp_peer(mx, from, maskport);
}

udp_mx_peer *bnet_forwarder::new_udp_client(bnet_udp_forwarder *mx, const socket_address& from)
{
	bnet_udp_peer *cli = static_cast<bnet_udp_peer *>(new_udp_peer(*external_udp, mx, from));
	masking_udp.insert(address_map::value_type(cli->via->get_my_address(), from));
	if (DEBUG(BNETUDPFLOW))
		logger	<< timestamp << " client map: ext " << cli->via->get_my_address() << " -> " << from << endl;
	return cli;
}

udp_mx_peer *bnet_forwarder::new_udp_gamer(bnet_udp_forwarder *mx, const socket_address& from)
{
	return new_udp_peer(*internal_udp, mx, from);
}

socket_address::ipportpair bnet_forwarder::map_host_for_remote(const socket_address::ipportpair& addr)
{
	udp_mx_map::iterator p = internal_udp->find(addr);
	if (p != internal_udp->end()) {	// host zewnętrzny zamaskowany przez nas
		if (DEBUG(CONNS))
			logger	<< timestamp << ' ' << addr << " mapped to " << p->second->get_forward_target()
				<< " (remote for remote)" << endl;
		return p->second->get_forward_target();
	}

	p = external_udp->find(addr);
	if (p != external_udp->end()) {	// host wewnętrzny niezamaskowany
		socket_address mask;
		p->second->get_my_address(mask);
		if (DEBUG(CONNS))
			logger	<< timestamp << ' ' << addr << " mapped to " << mask << " (local for remote)" << endl;
		return mask;
	}

#ifndef ALL_IPS_EXTERNAL
	logger	<< timestamp << " **BUG** " << addr << " not found in mappings; sending out unchanged" << endl;
#endif
	return addr;
}

socket_address::ipportpair bnet_forwarder::map_host_for_local(const socket_address::ipportpair& addr)
{
	address_map::iterator p = masking_udp.find(addr);

	if (p == masking_udp.end()) {	// host zewnętrzny
		socket_address mask;
		bnet_udp_forwarder *maskport = create_udp_masking_port(*internal_udp, addr);
		maskport->get_my_address(mask);
		if (DEBUG(CONNS))
			logger	<< timestamp << ' ' << addr << " mapped to " << mask << " (remote for local)" << endl;
		return mask;
	} else {			// host wewnętrzny
		if (DEBUG(CONNS))
			logger	<< timestamp << ' ' << addr << " mapped to " << p->second << " (local for local)" << endl;
#ifndef ALL_IPS_EXTERNAL
		return p->second;	// <- oryginalny adres klienta
#else
		return addr;
#endif
	}
}

socket_address::ipportpair bnet_forwarder::map_host_for_addr(const socket_address::ipportpair& addr,
		const socket_address::ipportpair& dest)
{
	bool local = (external_udp->find(dest) != external_udp->end());	// docelowy adres jest adresem klienta lokalnego

	if (DEBUG(CONNS))
		logger	<< timestamp << " mapping " << addr << " for " << dest << (local ? " (local)" : " (remote)") << endl;

	if (local)
		return map_host_for_local(addr);
	else	return map_host_for_remote(addr);
}

void bnet_forwarder::drop_udp_forwarding(bnet_udp_forwarder *fwd)
{
	masking_udp.erase(fwd->get_my_address());
	delete fwd;
}

