#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <iostream>
using namespace std;
#include "ioclass.hh"
#include "logger.hh"

/**************************************
 * io_class
 */

io_class::io_class(int _fd, bool _own)
	: event(_fd, 0), fd(_fd), own(_own)
{
	if (_fd < 0)
		throw int(errno);
}

io_class::~io_class()
{
	if (own)
		close(fd);
}

void io_class::handle(int _fd, int evflags)
{
	if (fd != _fd)
		abort();
	try {
		if (evflags & EV_TIMEOUT)
			poll_timedout();
		if (evflags & EV_READ)
			poll_read();
		if (evflags & EV_WRITE)
			poll_write();
	} catch (int) { }
}

/**************************************
 * io_socket
 */

static void setup_fd_or_die(int fd)
{
	int one = 1;
	if (::setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &one, sizeof(one)) < 0)
		; //throw int(errno);
	if (::fcntl(fd, F_SETFL, O_NONBLOCK) < 0) {
		int e = errno;
		::close(fd);
		throw int(e);
	}
}

io_socket::io_socket(int _fd, bool _own)
	: io_class(_fd, _own)
{
	setup_fd_or_die(_fd);
}

io_socket::io_socket(int domain, int type, int proto)
	: io_class(socket(domain, type, proto), true)
{
	setup_fd_or_die(fd);
}

void io_socket::bind(const socket_address& sa)
{
	if (DEBUG(CONNS))
		logger	<< timestamp << " bind [" << fd << "] to " << sa << endl;

	if (::bind(fd, sa.const_sockaddr(), sa.salen) < 0) {
		logger	<< timestamp << " bind [" << fd << "] failed: " << errno_str(errno) << endl;
		throw int(errno);
	}
}

void io_socket::get_my_address(socket_address& sa) const
{
	sa.salen = sa.salen_max();
	if (getsockname(fd, sa.sockaddr(), &sa.salen) < 0)
		throw int(errno);
}

void io_socket::get_peer_address(socket_address& sa) const
{
	sa.salen = sa.salen_max();
	if (getpeername(fd, sa.sockaddr(), &sa.salen) < 0)
		throw int(errno);
}

/**************************************
 * io_stream_listener
 */

io_stream_listener::io_stream_listener(int domain, int proto)
	: io_socket(domain, SOCK_STREAM, proto), listening(false)
{
}

io_stream_listener::~io_stream_listener()
{
	if (listening && DEBUG(CONNS))
		logger	<< timestamp << " TCP listener [" << fd << "] shut down" << endl;
}

void io_stream_listener::listen()
{
	if (::listen(fd, 4) < 0)
		throw int(errno);
	event::set(fd, EV_READ|EV_PERSIST);
	event::add(*this);
	listening = true;

	if (DEBUG(CONNS)) {
		socket_address me;
		get_my_address(me);
		logger	<< timestamp << " TCP listener [" << fd << "] started on " << me << endl;
	}
}

void io_stream_listener::poll_read()
{
	int nfd = ::accept(fd, NULL, NULL);
	if (nfd >= 0)
		poll_newconn(nfd);
	else if (errno != EINTR && errno != EAGAIN && errno != ECONNABORTED && errno != EPERM)
		throw int(errno);
}

/**************************************
 * io_stream_connector
 */

io_stream_connector::io_stream_connector(int domain, int proto)
	: io_socket(domain, SOCK_STREAM, proto)
{
}

void io_stream_connector::connect(const socket_address& sa, const timeout& to)
{
	if (DEBUG(CONNS))
		logger	<< timestamp << " connect [" << fd << "] to " << sa << endl;

	if (::connect(fd, sa.const_sockaddr(), sa.salen) < 0) {
		if (errno != EINPROGRESS)
			throw int(errno);
		else {
			event::set(EV_WRITE);
			event::add(*this, to);
		}
	} else
		poll_connected();
}

void io_stream_connector::poll_write()
{
	int err, errlen = sizeof(err);
	if (getsockopt(fd, SOL_SOCKET, SO_ERROR, &err, (socklen_t *)&errlen) < 0)
		throw int(errno);
	if (!err)
		poll_connected();
	else	poll_error(err);
}

timeout io_stream_connector::default_connect_timeout(20, 0);

/**************************************
 * address_detector
 */

void address_detector::execute(const socket_address& addr, socket_address& dest)
{
	connect(addr);
	get_my_address(dest);
	event::del(*this);
}

/**************************************
 * io_datagram_socket
 */

io_datagram_socket::io_datagram_socket(int domain, int proto)
	: io_socket(domain, SOCK_DGRAM, proto)
{
}

void io_datagram_socket::associate(const socket_address& peer)
{
	if (DEBUG(CONNS))
		logger	<< timestamp << " associate [" << fd << "] to " << peer << endl;

	if (::connect(fd, peer.const_sockaddr(), peer.salen) < 0) {
		logger	<< timestamp << " associate [" << fd << "] failed: " << errno_str(errno) << endl;
		throw int(errno);
	}
}

